# stdlib imports
import json
import os

# third-party imports
import mock

# local imports
from shpkpr.marathon_lb import get_marathon_lb_urls
from shpkpr.marathon_lb import prepare_deploy
from shpkpr.marathon_lb import parse_haproxy_stats
from shpkpr.marathon_lb import fetch_app_listeners
from shpkpr.marathon_lb import get_svnames_from_tasks
from shpkpr.marathon_lb import find_drained_task_ids


def _load_listeners():
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'fixtures', 'haproxy_stats.csv')
    with open(path) as f:
        return parse_haproxy_stats(f.read())


def _load_marathon_apps():
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'fixtures', 'bluegreen_apps.json')
    with open(path) as f:
        return json.loads(f.read())['apps']


def _load_blue_app():
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'fixtures', 'bluegreen_app_blue.json')
    with open(path) as f:
        return json.loads(f.read())['apps'][0]


@mock.patch('socket.gethostbyname_ex',
            mock.Mock(side_effect=lambda hostname:
                      (hostname, [], ['127.0.0.1', '127.0.0.2'])))
def test_get_marathon_lb_urls():
    marathon_lb_urls = get_marathon_lb_urls('http://foobar.com:9090')

    assert 'http://127.0.0.1:9090' in marathon_lb_urls
    assert 'http://127.0.0.2:9090' in marathon_lb_urls
    assert 'http://127.0.0.3:9090' not in marathon_lb_urls


def test_prepare_deploy_swap_colors():
    previous_apps = _load_marathon_apps()
    blue_app = _load_blue_app()

    deploy = prepare_deploy(previous_apps, blue_app)

    assert deploy['labels']['HAPROXY_DEPLOYMENT_COLOUR'] == 'green'
    assert deploy['instances'] == blue_app['instances']


def test_prepare_deploy_first_deploy():
    previous_apps = []
    blue_app = _load_blue_app()

    deploy = prepare_deploy(previous_apps, blue_app)

    assert deploy['labels']['HAPROXY_DEPLOYMENT_COLOUR'] == 'blue'
    assert deploy['instances'] == blue_app['instances']


def test_parse_haproxy_stats():
    results = _load_listeners()

    assert results[1].pxname == 'http-in'
    assert results[1].svname == 'IPv4-direct'

    assert results[2].pxname == 'http-out'
    assert results[2].svname == 'IPv4-cached'


@mock.patch('shpkpr.cli.logger.Logger')
@mock.patch('shpkpr.marathon_lb.fetch_combined_haproxy_stats',
            mock.Mock(side_effect=lambda _, __: _load_listeners()))
def test_fetch_app_listeners(logger):
    app = {
            'labels': {
              'HAPROXY_DEPLOYMENT_GROUP': 'foobar',
              'HAPROXY_0_PORT': '8080'
            }
          }

    app_listeners = fetch_app_listeners(logger, app, [])

    assert app_listeners[0].pxname == 'foobar_8080'
    assert len(app_listeners) == 1


def test_get_svnames_from_tasks():
    apps = _load_marathon_apps()
    tasks = apps[0]['tasks']

    task_svnames = get_svnames_from_tasks(tasks)

    assert '10_0_6_25_16916' in task_svnames
    assert '10_0_6_25_31184' in task_svnames


def test_find_drained_task_ids():
    listeners = _load_listeners()
    haproxy_instance_count = 2
    apps = _load_marathon_apps()
    app = apps[0]

    results = find_drained_task_ids(app, listeners, haproxy_instance_count)

    assert app['tasks'][0]['id'] in results  # 2 down, no sessions
    assert app['tasks'][1]['id'] not in results  # 1 up, one down
    assert app['tasks'][2]['id'] not in results  # 2 down, one w/ session



# third-party import
import click

# local imports
from shpkpr.cli import options
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.cli.logger import pass_logger


@click.command('show', short_help='Show application details.', context_settings=CONTEXT_SETTINGS)
@options.output_formatter
@options.application_id
@options.marathon_client
@pass_logger
def cli(logger, marathon_client, application_id, output_formatter):
    """Shows detailed information for one or more applications.
    """
    if application_id is None:
        payload = marathon_client.list_applications()
    else:
        payload = marathon_client.get_application(application_id)
    logger.log(output_formatter.format(payload))



# third-party imports
import mock
import responses


def test_no_args(runner):
    result = runner(['deploy'])

    assert result.exit_code == 2
    assert 'Usage:' in result.output


def test_help(runner):
    result = runner(['deploy', '--help'])

    assert result.exit_code == 0
    assert 'Usage:' in result.output
    assert 'Deploy application from template.' in result.output


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_no_force(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOY_DOMAIN': 'mydomain.com',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['deploy', '--template', _tmpl_path, 'RANDOM_LABEL=some_value'], env=env)

    assert result.exit_code == 0


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_force(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app?force=true',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOY_DOMAIN': 'mydomain.com',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['deploy', '--template', _tmpl_path, '--force', 'RANDOM_LABEL=some_value'], env=env)

    assert result.exit_code == 0


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_multiple_templates(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_APPLICATION_2': 'test-app-2',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOY_DOMAIN': 'mydomain.com',
        'SHPKPR_RANDOM_LABEL': 'some_value',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    _tmpl_path_2 = "tests/fixtures/templates/marathon/test-2.json.tmpl"
    result = runner(['deploy', '--template', _tmpl_path, '--template', _tmpl_path_2], env=env)

    assert result.exit_code == 0


@responses.activate
def test_dry_run(runner, json_fixture):
    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['deploy', '--dry-run', '--template', _tmpl_path, 'RANDOM_LABEL=some_value'], env=env)

    assert result.exit_code == 0
    assert '--dry-run' in result.output


@responses.activate
def test_dry_run_fail(runner, json_fixture):
    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['deploy', '--dry-run', '--template', _tmpl_path, 'RANDOM_LABEL=some_value'], env=env)

    assert not result.exit_code == 0



# stdlib imports
import sys

# third-party imports
import click


class Logger(object):

    def __init__(self):
        self.buffer = sys.stdout

    def log(self, msg, *args, **kwargs):
        """Logs a message to stdout."""
        if args:
            msg %= args
        click.echo(msg, file=self.buffer, **kwargs)

    def style(self, msg, *args, **kwargs):
        """Wrapper around click.style to allow modules to access this
        functionality via the context object without having to depend on
        click.
        """
        return click.style(msg, *args, **kwargs)


pass_logger = click.make_pass_decorator(Logger, ensure=True)



# stdlib imports
from collections import namedtuple
import csv
from datetime import datetime
import socket
import time

# third-party imports
import click
import requests
import six.moves.urllib as urllib

# local imports
from shpkpr import exceptions


# amount of time (in seconds) between subsequent polls of the load balancer
# during a blue/green deploy.
MARATHON_LB_POLL_INTERVAL = 5


class SwapApplicationTimeout(exceptions.ShpkprException):
    pass


def _get_alias_records(hostname):
    """Return all IPv4 A records for a given hostname
    """
    return socket.gethostbyname_ex(hostname)[2]


def _unparse_url_alias(url, addr):
    """Reassemble a url object into a string but with a new address
    """
    return urllib.parse.urlunparse((url[0],
                                    addr + ":" + str(url.port),
                                    url[2],
                                    url[3],
                                    url[4],
                                    url[5]))


def get_marathon_lb_urls(marathon_lb_url):
    """Return a list of urls for all Aliases of the
       marathon_lb url passed in as an argument
    """
    url = urllib.parse.urlparse(marathon_lb_url)
    addrs = _get_alias_records(url.hostname)
    return [_unparse_url_alias(url, addr) for addr in addrs]


def fetch_haproxy_pids(logger, haproxy_url):
    try:
        response = requests.get(haproxy_url + "/_haproxy_getpids")
        response.raise_for_status()
    except requests.exceptions.RequestException:
        logger.log("Caught exception when retrieving HAProxy"
                   " pids from " + haproxy_url)
        raise

    return response.text.split()


def check_haproxy_reloading(logger, haproxy_url):
    """Return False if haproxy has only one pid, it is not reloading.
       Return True if we catch an exception while making a request to
       haproxy or if more than one pid is returned
    """
    try:
        pids = fetch_haproxy_pids(logger, haproxy_url)
    except requests.exceptions.RequestException:
        # Assume reloading on any error, this should be caught with a timeout
        return True

    if len(pids) > 1:
        logger.log("Waiting for {} pids on {}".format(len(pids), haproxy_url))
        return True

    return False


def any_marathon_lb_reloading(logger, marathon_lb_urls):
    return any([check_haproxy_reloading(logger, url) for url in marathon_lb_urls])


def fetch_haproxy_stats(logger, haproxy_url):
    try:
        response = requests.get(haproxy_url + "/haproxy?stats;csv")
        response.raise_for_status()
    except requests.exceptions.RequestException:
        logger.log("Caught exception when retrieving HAProxy"
                   " stats from " + haproxy_url)
        raise

    return response.text


def fetch_combined_haproxy_stats(logger, marathon_lb_urls):
    s = ''.join([fetch_haproxy_stats(logger, url) for url in marathon_lb_urls])
    return parse_haproxy_stats(s)


def parse_haproxy_stats(csv_data):
    rows = csv_data.splitlines()
    headings = rows.pop(0).lstrip('# ').rstrip(',\n').split(',')
    csv_reader = csv.reader(rows, delimiter=',', quotechar="'")

    Row = namedtuple('Row', headings)

    return [Row(*row[0:-1]) for row in csv_reader if row[0][0] != '#']


def get_deployment_label(app):
    return get_deployment_group(app) + "_" + app['labels']['HAPROXY_0_PORT']


def _is_app_listener(app, listener):
    return (listener.pxname == get_deployment_label(app) and
            listener.svname not in ['BACKEND', 'FRONTEND'])


def fetch_app_listeners(logger, app, marathon_lb_urls):
    haproxy_stats = fetch_combined_haproxy_stats(logger, marathon_lb_urls)
    return [l for l in haproxy_stats if _is_app_listener(app, l)]


def waiting_for_listeners(new_app, old_app, listeners, haproxy_count):
    listener_count = (len(listeners) / haproxy_count)
    return listener_count != new_app['instances'] + old_app['instances']


def get_deployment_target(app):
    return int(app['labels']['HAPROXY_DEPLOYMENT_TARGET_INSTANCES'])


def waiting_for_up_listeners(app, listeners, haproxy_count):
    up_listeners = [l for l in listeners if l.status == 'UP']
    up_listener_count = (len(up_listeners) / haproxy_count)

    return up_listener_count < get_deployment_target(app)


def _has_pending_requests(listener):
    return int(listener.qcur or 0) > 0 or int(listener.scur or 0) > 0


def select_drained_listeners(listeners):
    draining_listeners = [l for l in listeners if l.status == 'MAINT']
    return [l for l in draining_listeners if not _has_pending_requests(l)]


def get_svnames_from_task(task):
    prefix = task['host'].replace('.', '_')
    for port in task['ports']:
        yield(prefix + '_{}'.format(port))


def get_svnames_from_tasks(tasks):
    svnames = []
    for task in tasks:
        svnames += get_svnames_from_task(task)
    return svnames


def find_drained_task_ids(app, listeners, haproxy_count):
    """Return app tasks which have all haproxy listeners down and drained
       of any pending sessions or connections
    """
    tasks = zip(get_svnames_from_tasks(app['tasks']), app['tasks'])
    drained_listeners = select_drained_listeners(listeners)

    drained_task_ids = []
    for svname, task in tasks:
        task_listeners = [l for l in drained_listeners if l.svname == svname]
        if len(task_listeners) == haproxy_count:
            drained_task_ids.append(task['id'])

    return drained_task_ids


def max_wait_exceeded(max_wait, timestamp):
    return (time.time() - timestamp > max_wait)


def check_time_and_sleep(max_wait, timestamp):
    if max_wait_exceeded(max_wait, timestamp):
        raise SwapApplicationTimeout('Max wait Time Exceeded')

    return time.sleep(MARATHON_LB_POLL_INTERVAL)


def swap_bluegreen_apps(logger, force, max_wait, marathon_client,
                        marathon_lb_url, new_app, old_app, timestamp):
    while True:

        check_time_and_sleep(max_wait, timestamp)

        old_app = marathon_client.get_application(old_app['id'])
        new_app = marathon_client.get_application(new_app['id'])

        logger.log("Existing app running {} instances, "
                   "new app running {} instances"
                   .format(old_app['instances'], new_app['instances']))

        marathon_lb_urls = get_marathon_lb_urls(marathon_lb_url)
        haproxy_count = len(marathon_lb_urls)

        try:
            listeners = fetch_app_listeners(logger, new_app, marathon_lb_urls)
        except requests.exceptions.RequestException:
            # Restart loop if we hit an exception while loading listeners,
            # this may be normal behaviour
            continue

        logger.log("Found {} app listeners across {} HAProxy instances"
                   .format(len(listeners), haproxy_count))

        if waiting_on_marathon_lb(logger, marathon_lb_urls, new_app, old_app,
                                  listeners, haproxy_count):
            continue

        drained_task_ids = \
            find_drained_task_ids(old_app, listeners, haproxy_count)

        if ready_to_delete_old_app(new_app, old_app, drained_task_ids):
            logger.log("About to delete old app {}".format(old_app['id']))
            return safe_delete_app(force, marathon_client, old_app)


def waiting_on_marathon_lb(logger, marathon_lb_urls, new_app, old_app,
                           listeners, haproxy_count):
    if any_marathon_lb_reloading(logger, marathon_lb_urls):
        return True

    if waiting_for_listeners(new_app, old_app, listeners, haproxy_count):
        return True

    if waiting_for_up_listeners(new_app, listeners, haproxy_count):
        return True

    if waiting_for_drained_listeners(listeners):
        return True

    return False


def ready_to_delete_old_app(new_app, old_app, drained_task_ids):
    return (int(new_app['instances']) == get_deployment_target(new_app) and
            len(drained_task_ids) == int(old_app['instances']))


def waiting_for_drained_listeners(listeners):
    return len(select_drained_listeners(listeners)) < 1


def safe_delete_app(force, marathon_client, app):
    if force or click.confirm("Continue?"):
        marathon_client.delete_application(app['id'])
        return True
    else:
        return False


def get_service_port(app):
    try:
        return \
            int(app['container']['docker']['portMappings'][0]['servicePort'])
    except KeyError:
        return int(app['ports'][0])


def set_service_port(app, servicePort):
    try:
        app['container']['docker']['portMappings'][0]['servicePort'] = \
          int(servicePort)
    except KeyError:
        app['ports'][0] = int(servicePort)

    return app


def validate_app(app):
    if app['id'] is None:
        raise Exception("App doesn't contain a valid App ID")
    if 'labels' not in app:
        raise Exception("No labels found. Please define the"
                        "HAPROXY_DEPLOYMENT_GROUP label")
    if 'HAPROXY_DEPLOYMENT_GROUP' not in app['labels']:
        raise Exception("Please define the "
                        "HAPROXY_DEPLOYMENT_GROUP label")
    if 'HAPROXY_DEPLOYMENT_ALT_PORT' not in app['labels']:
        raise Exception("Please define the "
                        "HAPROXY_DEPLOYMENT_ALT_PORT label")


def set_app_ids(app, colour):
    app['labels']['HAPROXY_APP_ID'] = app['id']
    app['id'] = app['id'] + '-' + colour

    return app


def set_service_ports(app, servicePort):
    app['labels']['HAPROXY_0_PORT'] = str(get_service_port(app))
    try:
        app['container']['docker']['portMappings'][0]['servicePort'] = \
          int(servicePort)
        return app
    except KeyError:
        app['ports'][0] = int(servicePort)
        return app


def select_next_port(app):
    alt_port = int(app['labels']['HAPROXY_DEPLOYMENT_ALT_PORT'])
    if int(app['ports'][0]) == alt_port:
        return int(app['labels']['HAPROXY_0_PORT'])
    else:
        return alt_port


def select_next_colour(app):
    if app['labels'].get('HAPROXY_DEPLOYMENT_COLOUR') == 'blue':
        return 'green'
    else:
        return 'blue'


def sort_deploys(apps):
    return sorted(apps, key=lambda a: a.get('labels', {})
                  .get('HAPROXY_DEPLOYMENT_STARTED_AT', '0'))


def select_last_deploy(apps):
    return sort_deploys(apps).pop()


def get_deployment_group(app):
    return app.get('labels', {}).get('HAPROXY_DEPLOYMENT_GROUP')


def fetch_previous_deploys(marathon_client, app):
    apps = marathon_client.list_applications()
    app_deployment_group = get_deployment_group(app)
    return [a for a in apps if get_deployment_group(a) == app_deployment_group]


def prepare_deploy(previous_deploys, app):
    """ Return a blue or a green version of `app` based on prexisting deploys
    """
    if len(previous_deploys) > 0:
        last_deploy = select_last_deploy(previous_deploys)
        next_colour = select_next_colour(last_deploy)
        next_port = select_next_port(last_deploy)
        deployment_target_instances = last_deploy['instances']
    else:
        next_colour = 'blue'
        next_port = get_service_port(app)
        deployment_target_instances = app['instances']

    app = set_app_ids(app, next_colour)
    app = set_service_ports(app, next_port)
    app['labels']['HAPROXY_DEPLOYMENT_TARGET_INSTANCES'] = \
        str(deployment_target_instances)
    app['labels']['HAPROXY_DEPLOYMENT_COLOUR'] = next_colour
    app['labels']['HAPROXY_DEPLOYMENT_STARTED_AT'] = datetime.now().isoformat()

    return app



# stdlib imports
import functools
import json
import os

# third-party imports
import pytest
from click.testing import CliRunner

# local imports
from shpkpr.cli.entrypoint import cli


@pytest.fixture
def runner():
    runner = CliRunner()
    return functools.partial(runner.invoke, cli)


@pytest.fixture
def json_fixture():
    def _json_fixture(name):
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir, "fixtures", name + ".json")
        with open(path, 'r') as f:
            fixture = json.loads(f.read())
        return fixture
    return _json_fixture



# stdlib imports
import json

# third-party imports
import click
import time

# local imports
from shpkpr import exceptions
from shpkpr.cli import arguments
from shpkpr.cli import options
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.cli.logger import pass_logger
from shpkpr.template import load_values_from_environment
from shpkpr.template import render_json_template
from shpkpr.marathon import DeploymentFailed
from shpkpr.marathon_lb import SwapApplicationTimeout
from shpkpr.marathon_lb import fetch_previous_deploys
from shpkpr.marathon_lb import prepare_deploy
from shpkpr.marathon_lb import select_last_deploy
from shpkpr.marathon_lb import swap_bluegreen_apps
from shpkpr.marathon_lb import validate_app


class DualStackAlreadyExists(exceptions.ShpkprException):
    """Raised if a deploy is attempt whilst both a blue and green stack
    already exist in Marathon
    """
    exit_code = 2

    def format_message(self):
        msg = "Both blue and green stacks detected on Marathon, please resolve "
        msg += "before deploying. This may mean that another deploy is in progress."
        return msg


@click.command('bluegreen', short_help='perform a blue-green deploy of the application',
               context_settings=CONTEXT_SETTINGS)
@arguments.env_pairs
@options.template_names
@options.template_path
@options.env_prefix
@options.marathon_client
@options.marathon_lb_url
@options.max_wait
@options.force
@pass_logger
def cli(logger, marathon_client, marathon_lb_url, max_wait, force, env_prefix,
        template_path, template_names, env_pairs):
    """Perform a blue/green deploy
    """
    values = load_values_from_environment(prefix=env_prefix, overrides=env_pairs)

    rendered_templates = []
    for template_name in template_names:
        rendered_templates.append(render_json_template(template_path, template_name, **values))

    for app in rendered_templates:
        validate_app(app)

        # fetch all previous deploys from marathon and abort if there is more
        # than one stack currently active.
        previous_deploys = fetch_previous_deploys(marathon_client, app)
        if len(previous_deploys) > 1:
            raise DualStackAlreadyExists("Both blue and green apps detected")
        # transform the app to be deployed to apply the correct labels and
        # ID-change that will allow marathon-lb to cut traffic over as necessary.
        new_app = prepare_deploy(previous_deploys, app)

        logger.log('Final App Definition:')
        logger.log(json.dumps(new_app, indent=4, sort_keys=True))
        if force or click.confirm("Continue with deployment?"):
            try:
                deploy_and_swap(marathon_client,
                                new_app,
                                previous_deploys,
                                logger,
                                force,
                                max_wait,
                                marathon_lb_url)
            except (DeploymentFailed, SwapApplicationTimeout):
                remove_new_stack(marathon_client, logger, app['id'], force)


def deploy_and_swap(marathon_client, new_app, previous_deploys, logger, force,
                    max_wait, marathon_lb_url):
    """Deploy a new application and swap traffic from the old one once complete.
    """
    marathon_client.deploy([new_app]).wait(timeout=max_wait)

    if len(previous_deploys) == 0:
        # This was the first deploy, nothing to swap
        return True
    else:
        # This is a standard blue/green deploy, swap new app with old
        old_app = select_last_deploy(previous_deploys)
        return swap_bluegreen_apps(logger,
                                   force,
                                   max_wait,
                                   marathon_client,
                                   marathon_lb_url,
                                   new_app,
                                   old_app,
                                   time.time())


def remove_new_stack(marathon_client, logger, app_id, force):
    """Removes the newly started stack after a deploy error
    """
    logger.log("Deployment failed: removing newly deployed stack")
    success = marathon_client.delete_application(app_id, force=force)
    if success:
        logger.log("Successfully removed newly deployed stack: %s", app_id)
    else:
        logger.log("Unable to remove newly deployed stack, manual intervention required: %s", app_id)
    raise DeploymentFailed("Deployment failed")



# stdlib imports
import copy

# third-party imports
import pytest


def _check_exits_zero(result):
    assert result.exit_code == 0


def _check_exits_non_zero(result):
    assert result.exit_code != 0


def _check_output_contains(result, match):
    assert match in result.output


def _check_output_does_not_contain(result, match):
    assert match not in result.output


@pytest.mark.integration
def test_help(runner, env):
    result = runner(["--help"], env=env)
    _check_exits_zero(result)
    _check_output_contains(result, "Usage:")


@pytest.mark.integration
def test_show(runner, env):
    result = runner(["show"], env=env)
    _check_exits_non_zero(result)
    _check_output_contains(result, "Unable to retrieve application details from marathon: does not exist.")


@pytest.mark.integration
def test_deploy(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(["deploy", "-t", _tmpl_path, "RANDOM_LABEL=some_value"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_show_again(runner, env):
    result = runner(["show"], env=env)
    _check_exits_zero(result)
    _check_output_contains(result, '"id": "/%s"' % env['SHPKPR_APPLICATION'])


@pytest.mark.integration
def test_cron_show(runner, env):
    result = runner(["cron", "show"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_set(runner, env):
    _tmpl_path = 'tests/fixtures/templates/chronos/test-chronos.json.tmpl'
    result = runner(["cron", "set", "--template", _tmpl_path, "CHRONOS_JOB_NAME=shpkpr-test-job"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_after_add(runner, env):
    result = runner(["cron", "show"], env=env)

    _check_output_contains(result, '"name": "shpkpr-test-job"')
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_run(runner, env):
    result = runner(["cron", "run", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_delete_tasks(runner, env):
    result = runner(["cron", "delete-tasks", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_delete(runner, env):
    result = runner(["cron", "delete", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_after_delete(runner, env):
    result = runner(["cron", "show"], env=env)

    _check_output_does_not_contain(result, '"name": "shpkpr-test-job"')
    _check_exits_zero(result)


@pytest.mark.integration
def test_bluegreen_deploy(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen.json.tmpl"
    result = runner(["bluegreen", "-t", _tmpl_path, "--force"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_deployed(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-blue"

    result = runner(["show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-blue")


@pytest.mark.integration
def test_bluegreen_deploy_with_existing_app(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen.json.tmpl"
    result = runner(["bluegreen", "-t", _tmpl_path, "--force"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_swapped(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-green"

    result = runner(["show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-green")



"""Option parameter definitions for shpkpr.

All option parameters should be defined in this module, commands should not
define their options directly, instead all options should be defined here and
imported as required. This helps maintain consistency where a single option is
used for multiple commands.
"""
# stdlib imports
import os

# third-party imports
import click

# local imports
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.cli.formatter import OutputFormatter
from shpkpr.marathon import MarathonClient


application_id = click.option(
    '-a', '--application',
    'application_id',
    envvar="{0}_APPLICATION".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    help="ID/name of the application to scale.",
)


dry_run = click.option(
    '--dry-run',
    is_flag=True,
    help='Enables dry-run mode. Shpkpr will not attempt to contact Marathon when this is enabled.',
)


env_prefix = click.option(
    '-e', '--env_prefix',
    'env_prefix',
    default=CONTEXT_SETTINGS['auto_envvar_prefix'],
    help="Prefix used to restrict environment vars used for templating.",
)


force = click.option(
    '--force',
    is_flag=True,
    help='Force update even if a deployment is in progress.',
)


marathon_client = click.option(
    '--marathon_url',
    'marathon_client',
    envvar="{0}_MARATHON_URL".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=True,
    help="URL of the Marathon API to use.",
    callback=lambda c, p, v: MarathonClient(v)
)


template_path = click.option(
    '--template_dir',
    'template_path',
    envvar="{0}_TEMPLATE_DIR".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    type=str,
    default=os.getcwd(),
    help="Base directory in which your templates are stored.",
)


template_names = click.option(
    '-t', '--template',
    'template_names',
    envvar="{0}_TEMPLATES".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    type=str,
    required=True,
    help="Path of the template to use for deployment.",
    multiple=True,
)

chronos_url = click.option(
    '--chronos_url',
    'chronos_url',
    envvar="{0}_CHRONOS_URL".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=True,
    help='URL of the Chronos endpoint to use'
)

chronos_version = click.option(
    '--chronos_version',
    'chronos_version',
    envvar="{0}_CHRONOS_VERSION".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=False,
    help='Verson of the Chronos endpoint to use',
    default='3.0.2'
)

job_name = click.option(
    '-j', '--job-name',
    type=str,
    help='Restrict command to specific job.',
)


marathon_lb_url = click.option(
    '--marathon_lb_url',
    'marathon_lb_url',
    envvar="{0}_MARATHON_LB_URL".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=True,
    help="url of marathon lb",
)


max_wait = click.option(
    '--max-wait',
    'max_wait',
    type=int,
    help='Maximum amount of time to wait for deployment to finish before aborting.',
    envvar="{0}_MAX_WAIT".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default=300,
)

output_formatter = click.option(
    '--output-format',
    'output_formatter',
    type=click.Choice(["json", "yaml"]),
    help='Serialization format to use when printing config data to stdout.',
    envvar="{0}_OUTPUT_FORMAT".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default="json",
    callback=lambda c, p, v: OutputFormatter(v)
)



# stdlib imports
from distutils.version import StrictVersion

# third-party imports
import click
import chronos

# local imports
from shpkpr.cli import arguments, options
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.cli.logger import pass_logger
from shpkpr.template import load_values_from_environment
from shpkpr.template import render_json_template


def chronos_connect(chronos_url, chronos_version):
    if StrictVersion(chronos_version) >= StrictVersion('3.0.0'):
        api_version = 'v1'
    else:
        api_version = None
    return chronos.connect(chronos_url, scheduler_api_version=api_version)


@click.group('cron', short_help='Manage Chronos Jobs', context_settings=CONTEXT_SETTINGS)
@pass_logger
def cli(logger):
    """Manage Chronos Jobs.
    """


@cli.command('show', short_help='List Chronos Jobs as json', context_settings=CONTEXT_SETTINGS)
@options.chronos_url
@options.chronos_version
@options.job_name
@options.output_formatter
@pass_logger
def show(logger, chronos_url, chronos_version, job_name, output_formatter):
    """List application configuration.
    """
    jobs = chronos_connect(chronos_url, chronos_version).list()

    if job_name is None:
        payload = jobs
    else:
        payload = _find_job(jobs, job_name)
    logger.log(output_formatter.format(payload))


@cli.command('set', short_help='Add or Update a Chronos Job', context_settings=CONTEXT_SETTINGS)
@arguments.env_pairs
@options.chronos_url
@options.chronos_version
@options.env_prefix
@options.template_names
@options.template_path
@options.env_prefix
def set(chronos_url, chronos_version, template_path, template_names, env_prefix, env_pairs):
    """Add or Update a job in chronos.
    """
    values = load_values_from_environment(prefix=env_prefix, overrides=env_pairs)
    current_jobs = chronos_connect(chronos_url, chronos_version).list()

    for template_name in template_names:
        rendered_template = render_json_template(template_path, template_name, **values)
        if _find_job(current_jobs, rendered_template['name']):
            chronos_connect(chronos_url, chronos_version).update(rendered_template)
        else:
            chronos_connect(chronos_url, chronos_version).add(rendered_template)


@cli.command('delete', short_help='Deletes a Job from Chronos', context_settings=CONTEXT_SETTINGS)
@arguments.job_name
@options.chronos_url
@options.chronos_version
def delete(chronos_url, chronos_version, job_name):
    chronos_connect(chronos_url, chronos_version).delete(job_name)


@cli.command('delete-tasks', short_help='Terminate all tasks for a Chronos Job.', context_settings=CONTEXT_SETTINGS)
@arguments.job_name
@options.chronos_url
@options.chronos_version
def delete_tasks(chronos_url, chronos_version, job_name):
    chronos_connect(chronos_url, chronos_version).delete_tasks(job_name)


@cli.command('run', short_help='Runs a Chronos Job', context_settings=CONTEXT_SETTINGS)
@arguments.job_name
@options.chronos_url
@options.chronos_version
def run(chronos_url, chronos_version, job_name):
    chronos_connect(chronos_url, chronos_version).run(job_name)


def _find_job(jobs, job_name):
    return list(filter(lambda j: job_name == j["name"], jobs))



# third-party imports
import click

# local imports
from shpkpr.cli import arguments
from shpkpr.cli import options
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.cli.logger import pass_logger
from shpkpr.template import load_values_from_environment
from shpkpr.template import render_json_template


@click.command('deploy', short_help='Deploy application from template.', context_settings=CONTEXT_SETTINGS)
@arguments.env_pairs
@options.force
@options.dry_run
@options.template_names
@options.template_path
@options.env_prefix
@options.marathon_client
@pass_logger
def cli(logger, marathon_client, env_prefix, template_path, template_names, dry_run, force, env_pairs):
    """Deploy application from template.
    """
    # set dry_run param if the user has requested it. This happens in the
    # command callback as it only applies to the deploy command and doesn't make
    # sense anywhere else.
    marathon_client.dry_run = dry_run
    # read and render deploy template using values from the environment
    values = load_values_from_environment(prefix=env_prefix, overrides=env_pairs)
    rendered_templates = []
    for template_name in template_names:
        rendered_templates.append(render_json_template(template_path, template_name, **values))

    marathon_client.deploy(rendered_templates, force=force).wait()
